class_name DotColoredTexture2D extends Texture2D

func _init(pcolor: Color) -> void:
	color = pcolor

func _get_height() -> int:
	return size

func _get_width() -> int:
	return size

const size := 20
const shape := Vector2(size, size)
const radius := 6

@export var color := Color.WHITE

func _draw(to_canvas_item: RID, pos: Vector2, modulate: Color, _transpose: bool) -> void:
	var circle = range(0, 16).map(
		func (i):
			var r = (TAU / 16) * i
			return Vector2(cos(r) * 0.1, sin(r) * 0.1) + shape / 2 + pos
	)
	circle += [circle[0]]
	var colors = range(0, 16)
	colors.fill(color * modulate)
	colors += [colors[0]]
	RenderingServer.canvas_item_add_polyline(to_canvas_item, circle, colors, radius * 2, true)

func _draw_rect(to_canvas_item: RID, rect: Rect2, _tile: bool, modulate: Color, _transpose: bool) -> void:
	var circle = range(0, 16).map(
		func (i):
			var r = (TAU / 16) * i
			return Vector2(cos(r) * 0.1, sin(r) * 0.1) + rect.size / 2 + rect.position
	)
	circle += [circle[0]]
	var colors = range(0, 16)
	colors.fill(color * modulate)
	colors += [colors[0]]
	RenderingServer.canvas_item_add_polyline(to_canvas_item, circle, colors, radius * 2 * size / rect.size.y, true)
