class_name Timeline extends Control

var view_position := Vector2():
	set(value):
		view_position = value
		queue_redraw()
@export var title_box_style: StyleBox = StyleBoxEmpty.new()
@export var title_hover_box_style: StyleBox = StyleBoxEmpty.new()
@export var title_active_box_style: StyleBox = StyleBoxEmpty.new()
@export var title_font_sizes: Array[int] = [24, 20, 18]
@export var title_font_size_default := 15
@export var font_normal: Font = get_theme_default_font()
@export var font_bold: Font = get_theme_default_font()
@export var font_italic: Font = get_theme_default_font()
@export var font_bold_italic: Font = get_theme_default_font()
@export var colors := {
	&"default": Color.hex(0xffffffff),
	&"info": Color.hex(0x44bbffff),
	&"comment": Color.hex(0x888888ff),
	&"important": Color.hex(0xff8800ff),
	&"mark": Color.hex(0xffff00ff),
}
@export var scroll_speed := 20
@export var pan_speed := 2.0

const column_gap := 0
const row_gap := 4
const cap_sizes := [10, 9, 7, 7]
const cap_size_default := 4
const cap_gap := 2
var font_ascent := font_normal.get_ascent(get_theme_default_font_size())
var units := []
@onready var edit_view := $TextEdit

func get_title_font(bold := false, italic := false) -> Font:
	if bold:
		if italic:
			return font_bold_italic
		else:
			return font_bold
	else:
		if italic:
			return font_italic
		else:
			return font_normal
func get_title_font_size(level := 0) -> int:
	if level < title_font_sizes.size():
		return title_font_sizes[level]
	else:
		return title_font_size_default
func get_title_box() -> StyleBox:
	return title_box_style
func get_color(id, opacity) -> Color:
	var color: Color = colors.get(id, Color.from_string(id, Color.RED))
	color.a = opacity
	return color
func get_cap_size(level := 0) -> int:
	return cap_sizes[level] if level < cap_sizes.size() else cap_size_default
func get_cap_width(level := 0) -> int:
	return get_cap_size(level) * 2 + cap_gap
func calc_position(pos: Vector2) -> Vector2:
	return pos - view_position + get_rect().size / 2
func calc_rev_position(pos: Vector2) -> Vector2:
	return pos + view_position - get_rect().size / 2

func draw_cap(level: int, local_pos: Vector2, color := Color.WHITE) -> void:
	var pos := calc_position(local_pos)
	match level:
		0:
			draw_arc(pos, 1, 0, TAU, 16, color, 6, true)
			draw_arc(pos, 8, 0, TAU, 16, color, 1, true)
		1:
			draw_arc(pos, 1, 0, TAU, 16, color, 4, true)
			draw_arc(pos, 7, 0, TAU, 16, color, 1, true)
		2:
			draw_arc(pos, 1, 0, TAU, 16, color, 1, true)
			draw_arc(pos, 5, 0, TAU, 16, color, 1, true)
		3:
			draw_arc(pos, 1, 0, TAU, 16, color, 8, true)
		_:
			draw_arc(pos, 1, 0, TAU, 16, color, 4, true)
@warning_ignore("shadowed_variable_base_class")
func draw_title_box(local_pos: Vector2, size: Vector2, hover := false, active := false) -> void:
	var pos := calc_position(local_pos)
	var box := title_active_box_style if active else title_hover_box_style if hover else title_box_style
	box.draw(get_canvas_item(), Rect2(pos, size))
func draw_text_paragraph(local_pos: Vector2, text: TextParagraph, color := Color.WHITE) -> void:
	var pos := calc_position(local_pos)
	text.draw(get_canvas_item(), pos, color)
func draw_connect(src_pos: Vector2, rstart_pos: Vector2, rend_pos: Vector2, tar_pos: Vector2, color := Color.WHITE) -> void:
	src_pos = calc_position(src_pos)
	rstart_pos = calc_position(rstart_pos)
	rend_pos = calc_position(rend_pos)
	tar_pos = calc_position(tar_pos)
	var line := []
	if src_pos.x == tar_pos.x:
		line = [src_pos, tar_pos]
	else:
		line = [src_pos]
		line.append_array(range(rstart_pos.y, rend_pos.y).map(
			func (y):
				return Vector2(remap(ease(remap(y, rstart_pos.y, rend_pos.y, 0, 1), -3.0), 0, 1, rstart_pos.x, rend_pos.x), y)))
		line.append_array([rend_pos, tar_pos])
	draw_polyline(line, color, 0.8, true)

func _draw() -> void:
	for unit in units:
		unit.calc_shape(self)
	for unit in units:
		unit.draw_connect(self)
	for unit in units:
		unit.draw_body(self)

var c_hovered_unit = null
var c_actived_unit = null
func exchange_hovered_unit(unit) -> void:
	if c_hovered_unit != unit:
		if c_hovered_unit:
			c_hovered_unit.hover = false
		if unit:
			unit.hover = true
		c_hovered_unit = unit
		queue_redraw()
signal change_actived_unit(unit: TUnit)
func exchange_actived_unit(unit) -> void:
	if edit_view.visible:
		do_title_confirm()
	if c_actived_unit != unit:
		if c_actived_unit:
			c_actived_unit.active = false
		if unit:
			unit.active = true
		c_actived_unit = unit
		queue_redraw()
	change_actived_unit.emit(unit)

var c_mouse_main_pos := Vector2()
var c_actived_unit_pos := Vector2()
var c_mouse_mid_pos := Vector2()
func _gui_input(event: InputEvent) -> void:
	if event is InputEventMouseButton:
		if event.button_index == MOUSE_BUTTON_LEFT and event.pressed:
			c_mouse_main_pos = event.position
			var cur_actived_unit = null
			var mpos: Vector2 = calc_rev_position(event.position)
			for i in range(units.size() - 1, -1, -1):
				var unit = units[i]
				if unit.get_head_rect(self).has_point(mpos):
					accept_event()
					cur_actived_unit = unit
					break
			exchange_actived_unit(cur_actived_unit)
			if c_actived_unit:
				c_actived_unit_pos = Vector2(c_actived_unit.offset, c_actived_unit.delta)
			if event.double_click and cur_actived_unit:
				do_edit_unit(cur_actived_unit)
		if event.button_index == MOUSE_BUTTON_MIDDLE:
			accept_event()
			if event.pressed:
				c_mouse_mid_pos = event.position
				Input.mouse_mode = Input.MOUSE_MODE_CAPTURED
			else:
				Input.mouse_mode = Input.MOUSE_MODE_VISIBLE
				Input.warp_mouse(c_mouse_mid_pos)
		if event.button_index == MOUSE_BUTTON_WHEEL_DOWN:
			accept_event()
			view_position.y += scroll_speed
		if event.button_index == MOUSE_BUTTON_WHEEL_UP:
			accept_event()
			view_position.y -= scroll_speed
	if event is InputEventMouseMotion:
		if event.button_mask & MOUSE_BUTTON_MASK_LEFT:
			if c_actived_unit:
				accept_event()
				var pos = event.position
				var mvt = pos - c_mouse_main_pos
				if abs(mvt.y) <= 3:
					c_actived_unit.delta = c_actived_unit_pos.y
				else:
					c_actived_unit.delta = max(0, c_actived_unit_pos.y + mvt.y)
				if c_actived_unit.aligned:
					if abs(mvt.x) <= 12:
						c_actived_unit.offset = c_actived_unit_pos.x
					else:
						c_actived_unit.offset = c_actived_unit_pos.x + mvt.x
				queue_redraw()
		elif event.button_mask & MOUSE_BUTTON_MASK_MIDDLE:
			accept_event()
			view_position -= event.relative * pan_speed
		else:
			var cur_hovered_unit = null
			var mpos: Vector2 = calc_rev_position(event.position)
			for i in range(units.size() - 1, -1, -1):
				var unit = units[i]
				if unit.get_head_rect(self).has_point(mpos):
					accept_event()
					cur_hovered_unit = unit
					break
			exchange_hovered_unit(cur_hovered_unit)

func do_unit_previous() -> void:
	if c_actived_unit:
		var src = c_actived_unit.get_src()
		if src:
			exchange_actived_unit(src)
func do_unit_next() -> void:
	if c_actived_unit:
		for u in c_actived_unit.get_links():
			if u.aligned:
				exchange_actived_unit(u)
func do_unit_append() -> void:
	if c_actived_unit:
		var u = c_actived_unit.sub({&"title": "", &"aligned": false})
		units.append(u)
		do_edit_unit(u)
		u.calc_shape(self)
		queue_redraw()
func do_unit_append_aligned() -> void:
	if c_actived_unit:
		var u = c_actived_unit.sub({&"title": ""})
		units.append(u)
		do_edit_unit(u)
		u.calc_shape(self)
		queue_redraw()
func do_unit_duplicate() -> void:
	if c_actived_unit:
		var src = c_actived_unit.get_src()
		if src:
			var u = c_actived_unit.clone()
			src.append(u)
			units.append(u)
			do_edit_unit(u)
			u.calc_shape(self)
			queue_redraw()
func do_unit_remove() -> void:
	if c_actived_unit:
		var src = c_actived_unit.get_src()
		if src:
			c_actived_unit.delete(self)
			exchange_actived_unit(null)
			src.calc_shape(self)
			queue_redraw()
func do_unit_set_h(level: int) -> void:
	if c_actived_unit:
		c_actived_unit.level = level
		queue_redraw()
		exchange_actived_unit(c_actived_unit)
func do_unit_toggle_font_bold() -> void:
	if c_actived_unit:
		c_actived_unit.font_bold = not c_actived_unit.font_bold
		queue_redraw()
		exchange_actived_unit(c_actived_unit)
func do_unit_toggle_font_italic() -> void:
	if c_actived_unit:
		c_actived_unit.font_italic = not c_actived_unit.font_italic
		queue_redraw()
		exchange_actived_unit(c_actived_unit)
func do_unit_realign() -> void:
	if c_actived_unit:
		c_actived_unit.offset = 0
		queue_redraw()
		exchange_actived_unit(c_actived_unit)
func do_unit_set_color(color) -> void:
	if c_actived_unit:
		c_actived_unit.color = color
		queue_redraw()
		exchange_actived_unit(c_actived_unit)

var INPUT_MAP := {
	&"unit_previous": func (_event: InputEvent) -> void:
		if c_actived_unit:
			accept_event()
			do_unit_previous(),
	&"unit_next": func (_event: InputEvent) -> void:
		if c_actived_unit:
			accept_event()
			do_unit_next(),
	&"unit_append": func (_event: InputEvent) -> void:
		if c_actived_unit:
			accept_event()
			do_unit_append(),
	&"unit_append_aligned": func (_event: InputEvent) -> void:
		if c_actived_unit:
			accept_event()
			do_unit_append_aligned(),
	&"unit_duplicate": func (_event: InputEvent) -> void:
		if c_actived_unit:
			accept_event()
			do_unit_duplicate(),
	&"unit_remove": func (_event: InputEvent) -> void:
		if c_actived_unit:
			accept_event()
			do_unit_remove(),
}
func _input(event: InputEvent) -> void:
	for id in INPUT_MAP:
		if event.is_action_pressed(id):
			INPUT_MAP[id].call(event)
			break
func do_edit_unit(unit: TUnit) -> void:
	exchange_actived_unit(unit)
	edit_view.visible = true
	edit_view.position = calc_position(c_actived_unit.get_title_pos(self) + Vector2(0, c_actived_unit.get_head_shape(self).y + 4))
	edit_view.text = c_actived_unit.title
	edit_view.grab_focus()
	edit_view.select_all()
func do_title_confirm() -> void:
	if c_actived_unit:
		c_actived_unit.title = edit_view.text if edit_view.text else ""
		edit_view.visible = false
		c_actived_unit.calc_shape(self)
		queue_redraw()

func _on_text_edit_focus_exited() -> void:
	do_title_confirm()

func _on_text_edit_text_submitted(_new_text: String) -> void:
	do_title_confirm()


class TUnit:
	var title := "":
		set(value):
			title = value
			clear_cache()
	var detail := ""
	var level := 0:
		set(value):
			level = value
			clear_cache()
	var color = "default"
	var font_bold := false:
		set(value):
			font_bold = value
			clear_cache()
	var font_italic := false:
		set(value):
			font_italic = value
			clear_cache()
	var delta := 0:
		set(value):
			delta = value
			clear_cache()
	var offset := 0:
		set(value):
			offset = value
			clear_cache()
	var aligned := false:
		set(value):
			aligned = value
			clear_cache()
	
	var hover := false
	var active := false
	var src = null
	var links := []
	func get_links() -> Array:
		var res = links.map(func (ref): return ref.get_ref())
		return res
	func get_src() -> TUnit:
		return src.get_ref() if src is WeakRef else null
	func set_src(unit: TUnit) -> void:
		var pre_src = get_src()
		if pre_src != null:
			pre_src.remove(self)
		src = weakref(unit)
	
	func _init(params := {}) -> void:
		title = params.get(&"title", "")
		detail = params.get(&"detail", "")
		level = params.get(&"level", 0)
		color = params.get(&"color", "default")
		font_bold = params.get(&"font_bold", false)
		font_italic = params.get(&"font_italic", false)
		delta = params.get(&"delta", 0)
		offset = params.get(&"offset", 0)
		title = params.get(&"title", "")
		aligned = params.get(&"aligned", true)
	
	@warning_ignore("incompatible_ternary")
	func to_data() -> Dictionary:
		var vsrc := get_src()
		var vlinks := get_links()
		
		return {
			&"title": title,
			&"detail": detail,
			&"level": level,
			&"color": color,
			&"font_bold": font_bold,
			&"font_italic": font_italic,
			&"delta": delta,
			&"offset": offset,
			&"aligned": aligned,
			
			&"uid": get_instance_id(),
			&"src": vsrc.get_instance_id() if vsrc else null,
			&"links": vlinks.map(func (link): return link.get_instance_id()),
		}
	
	func append(unit: TUnit) -> void:
		links.append(weakref(unit))
		unit.set_src(self)
		unit.clear_cache()
		clear_cache()
	func remove(unit: TUnit) -> void:
		links = links.filter(func (ref): return ref.get_ref() != unit)
		clear_cache()
	func delete(cav: Timeline) -> void:
		var vsrc := get_src()
		if vsrc:
			cav.units.erase(self)
			vsrc.remove(self)
			for unit in get_links():
				unit.delete(cav)
	func sub(params = {}) -> TUnit:
		params.merge({
			&"title": title,
			&"level": level,
			&"color": color,
			&"font_bold": font_bold,
			&"font_italic": font_italic,
		})
		var unit = TUnit.new(params)
		append(unit)
		return unit
	
	var pc_shape = null
	var c_shape = null
	var c_pos_offset = null
	var c_pos = null
	var c_title_text = null
	func get_shape(cav: Timeline) -> Vector2:
		if c_shape != null:
			return c_shape
		else:
			calc_shape(cav)
			return c_shape
	func get_pos_offset(cav: Timeline) -> Vector2:
		if c_pos_offset != null:
			return c_pos_offset
		else:
			calc_shape(cav)
			return c_pos_offset
	func get_head_shape(cav: Timeline) -> Vector2:
		var ts := get_title_shape(cav)
		return Vector2(ts.x + cav.get_cap_width(level), max(ts.y, cav.get_cap_size(level) * 2))
	func get_pos(cav: Timeline) -> Vector2:
		var vsrc := get_src()
		if c_pos != null:
			return c_pos + (vsrc.get_pos(cav) if vsrc != null else Vector2())
		else:
			if vsrc != null:
				vsrc.calc_shape(cav)
				return c_pos + vsrc.get_pos(cav)
			else:
				c_pos = Vector2()
				return c_pos
	func get_title_pos(cav: Timeline) -> Vector2:
		var hs := get_head_shape(cav)
		return get_pos(cav) + Vector2(cav.get_cap_width(level) - cav.get_cap_size(level), -hs.y / 2)
	func get_title_text(cav: Timeline) -> TextParagraph:
		if c_title_text != null:
			return c_title_text
		else:
			c_title_text = TextParagraph.new()
			c_title_text.add_string(title, cav.get_title_font(font_bold, font_italic), cav.get_title_font_size(level))
			return c_title_text
	func get_title_shape(cav: Timeline) -> Vector2:
		var box := cav.get_title_box()
		return get_title_text(cav).get_size() + Vector2(
			box.get_margin(SIDE_LEFT) + box.get_margin(SIDE_RIGHT),
			box.get_margin(SIDE_TOP) + box.get_margin(SIDE_BOTTOM),
		)
	func get_head_rect(cav: Timeline) -> Rect2:
		var hs := get_head_shape(cav)
		return Rect2(Vector2(-cav.get_cap_size(level), -hs.y / 2) + get_pos(cav), hs)
	func get_color(cav: Timeline, pactive := active) -> Color:
		return cav.get_color(color, 1.0 if pactive else 1.0 if hover else 0.5)
	func get_cap_offset(cav: Timeline) -> Vector2:
		return Vector2(cav.get_cap_size(level), get_head_shape(cav).y / 2)
	
	func draw_body(cav: Timeline) -> void:
		draw_cap(cav)
		draw_title(cav)
	func draw_cap(cav: Timeline) -> void:
		cav.draw_cap(level, get_pos(cav), get_color(cav, true))
	func draw_title(cav: Timeline) -> void:
		var pos := get_title_pos(cav)
		var shape := get_title_shape(cav)
		var inner_pos := pos + cav.get_title_box().get_offset()
		cav.draw_title_box(pos, shape, hover, active)
		cav.draw_text_paragraph(inner_pos, get_title_text(cav), get_color(cav))
	func draw_connect(cav: Timeline) -> void:
		var vsrc := get_src()
		if vsrc != null:
			var spos := vsrc.get_pos(cav)
			var tpos := get_pos(cav)
			var s_end := spos.y + vsrc.get_head_shape(cav).y / 2
			var t_start := tpos.y - get_head_shape(cav).y / 2
			cav.draw_connect(
				spos,
				Vector2(spos.x, max(s_end, t_start - 32)),
				Vector2(tpos.x, t_start),
				tpos, vsrc.get_color(cav))
	
	func call_reshape(cav: Timeline) -> void:
		var vsrc = get_src()
		if vsrc != null:
			vsrc.clear_cache()
			vsrc.calc_shape(cav)
	
	func calc_shape(cav: Timeline) -> void:
		var units := get_links()
		var width := 0
		var height := 0
		var h_size := get_head_shape(cav)
		var c_offset := get_cap_offset(cav)
		var axis_min := -c_offset.x
		var axis_max := h_size.x - c_offset.x
		for unit in units:
			var s = unit.get_shape(cav)
			if not unit.aligned:
				width += s.x + cav.column_gap
			height = max(height, s.y)
		width -= cav.column_gap
		var axis := -cav.get_cap_size(level) + h_size.x / 2 - width / 2.0
		var axis_y := h_size.y / 2 + cav.row_gap
		axis_min = min(axis_min, axis)
		axis_max = max(axis_max, axis + width)
		for unit in units:
			if unit.aligned:
				var cap_oft: Vector2 = unit.get_cap_offset(cav)
				unit.c_pos = Vector2(unit.offset, axis_y + unit.delta + cap_oft.y)
				axis_min = min(axis_min, unit.offset - unit.get_pos_offset(cav).x - cap_oft.x)
				axis_max = max(axis_max, unit.offset - unit.get_pos_offset(cav).x - cap_oft.x + unit.get_shape(cav).x)
			else:
				unit.c_pos = Vector2(axis, axis_y + unit.delta) + unit.get_cap_offset(cav) + unit.get_pos_offset(cav)
				axis += unit.get_shape(cav).x + cav.column_gap
		c_shape = Vector2(axis_max - axis_min, height + cav.row_gap + h_size.y)
		c_pos_offset = Vector2(-axis_min - cav.get_cap_size(level), 0)
		if c_shape != pc_shape:
			call_reshape(cav)
	
	func clear_cache() -> void:
		pc_shape = c_shape
		c_shape = null
		c_pos_offset = null
		c_pos = null
		c_title_text = null

func get_data() -> Array:
	return units.map(func (u): return u.to_data())

func get_json_data() -> String:
	return JSON.stringify(get_data())

func set_data(data: Array) -> void:
	var units_dict := {}
	for entry in data:
		var u = TUnit.new(entry)
		u.set_meta(&"raw_data", entry)
		units_dict[entry.uid] = u
	for uid in units_dict:
		var u = units_dict[uid]
		var entry = u.get_meta(&"raw_data")
		if entry.src != null:
			u.src = weakref(units_dict[entry.src])
		u.links = entry.links.map(func (l): return weakref(units_dict[l]))
	units = units_dict.values()
	queue_redraw()

func set_json_data(json: String) -> void:
	set_data(JSON.parse_string(json))

func init() -> void:
	units = [TUnit.new({&"title": "根节点"})]
	queue_redraw()
