extends Panel

@onready var tl_colors: Dictionary = %Timeline.colors

var file_path: String = "":
	set(value):
		file_path = value
		if is_empty():
			show_welcome()
			get_window().title = "Timeling"
		else:
			hide_welcome()
			get_window().title = file_path.get_basename().get_file().to_pascal_case() + " (" + file_path.get_file() + ')'

func _ready() -> void:
	get_recent_file_paths()
	var args = Array(OS.get_cmdline_args())
	match args:
		[_, "--new"]:
			new_file_in_cur()
		[_, var path]:
			open_file_in_cur(path)
		_:
			%WelcomeLayer.visible = true
	# FIXME
	for id in tl_colors:
		%Action_unit_set_color.add_icon_item(DotColoredTexture2D.new(tl_colors[id]), id.to_pascal_case())

func _on_timeline_change_actived_unit(unit) -> void:
	var is_disable = unit == null
	%Action_unit_append_aligned.disabled = is_disable
	%Action_unit_append.disabled = is_disable
	%Action_unit_duplicate.disabled = is_disable || unit.get_src() == null
	%Action_unit_remove.disabled = is_disable || unit.get_src() == null
	%Action_unit_set_h.disabled = is_disable
	%Action_unit_set_bold.disabled = is_disable
	%Action_unit_set_italic.disabled = is_disable
	%Action_unit_set_color.disabled = is_disable
	%Action_unit_realign.disabled = is_disable || unit.get_src() == null || not unit.aligned
	if unit:
		%Action_unit_set_h.select(unit.level)
		%Action_unit_set_bold.button_pressed = unit.font_bold
		%Action_unit_set_italic.button_pressed = unit.font_italic
		%Action_unit_set_color.select(tl_colors.keys().find(unit.color))


func _on_action_unit_append_aligned_pressed() -> void:
	%Timeline.do_unit_append_aligned()

func _on_action_unit_append_pressed() -> void:
	%Timeline.do_unit_append()

func _on_action_unit_duplicate_pressed() -> void:
	%Timeline.do_unit_duplicate()

func _on_action_unit_remove_pressed() -> void:
	%Timeline.do_unit_remove()

func _on_action_unit_set_h_item_selected(index: int) -> void:
	%Timeline.do_unit_set_h(index)

func _on_action_unit_set_bold_pressed() -> void:
	%Timeline.do_unit_toggle_font_bold()

func _on_action_unit_set_italic_pressed() -> void:
	%Timeline.do_unit_toggle_font_italic()

func _on_action_unit_realign_pressed() -> void:
	%Timeline.do_unit_realign()

func _on_action_unit_set_color_item_selected(index: int) -> void:
	%Timeline.do_unit_set_color(tl_colors.keys()[index])

func do_open_file() -> void:
	%OpenFile.visible = true

func do_save_to_file() -> void:
	%SaveFile.visible = true

func save_file() -> void:
	if file_path:
		save_to_file(file_path)
	else:
		%SaveFile.visible = true

func save_to_file(path: String) -> void:
	var file = FileAccess.open(path, FileAccess.WRITE)
	file.store_string(%Timeline.get_json_data())
	file.close()
	file_path = path
	if file_path:
		append_recent_file_path(file_path)

func open_file(path: String) -> void:
	if is_empty():
		open_file_in_cur(path)
	else:
		match await %OpenSelection.select():
			&"new":
				OS.create_instance([OS.get_executable_path(), path])
			&"current":
				open_file_in_cur(path)

func open_file_in_cur(path: String) -> void:
	var file = FileAccess.open(path, FileAccess.READ)
	%Timeline.set_json_data(file.get_as_text(true))
	file.close()
	file_path = path
	if file_path:
		append_recent_file_path(file_path)

func new_file() -> void:
	if is_empty():
		new_file_in_cur()
	else:
		match await %OpenSelection.select():
			&"new":
				OS.create_instance([OS.get_executable_path(), "--new"])
			&"current":
				new_file_in_cur()

func new_file_in_cur() -> void:
	file_path = ""
	%Timeline.init()
	if not is_empty():
		hide_welcome()

func is_empty() -> bool:
	return file_path == "" and %Timeline.units.size() == 0

func hide_welcome() -> void:
	%WelcomeLayer.visible = false

func show_welcome() -> void:
	%WelcomeLayer.visible = true

var recent_file_paths = null
signal recent_file_paths_changed

func get_recent_file_paths() -> Array:
	if recent_file_paths != null:
		return recent_file_paths
	else:
		var data := FileAccess.get_file_as_string("user://recent_files")
		var lines := Array(data.split('\n', false))
		var paths = lines.map(func(l): return l.strip_edges()).filter(func(p): return p != "")
		recent_file_paths = paths
		recent_file_paths_changed.emit()
		return recent_file_paths

func set_recent_file_paths(paths: Array) -> void:
	var opt_file := FileAccess.open("user://recent_files", FileAccess.WRITE)
	for path in paths:
		opt_file.store_line(path)
	opt_file.close()
	recent_file_paths = paths
	recent_file_paths_changed.emit()

func remove_recent_file_path(path: String) -> void:
	set_recent_file_paths(recent_file_paths.filter(func(p): return p != path))

func clear_recent_file_paths() -> void:
	set_recent_file_paths([])

func append_recent_file_path(path: String) -> void:
	var index = recent_file_paths.find(path)
	if index >= 0:
		recent_file_paths.remove_at(index)
	recent_file_paths.push_front(path)
	set_recent_file_paths(recent_file_paths)
