extends Panel

@onready var viewer = get_tree().get_first_node_in_group(&"viewer")
var recent_file_paths = []:
	set(value):
		recent_file_paths = value
		%RecentList.clear()
		for path in recent_file_paths:
			var vname = path.get_basename().get_file().to_pascal_case()
			%RecentList.add_item(vname + " (" + path + ')')

func _on_new_file_pressed() -> void:
	await viewer.new_file()

func _on_open_file_pressed() -> void:
	await viewer.do_open_file()

func _on_recent_list_item_activated(index: int) -> void:
	viewer.open_file(recent_file_paths[index])

func update_recent_file_paths() -> void:
	recent_file_paths = viewer.get_recent_file_paths()

func _on_clear_recent_pressed() -> void:
	viewer.clear_recent_file_paths()

func _on_delete_recent_pressed() -> void:
	var selects := Array(%RecentList.get_selected_items())
	if selects.size():
		viewer.remove_recent_file_path(recent_file_paths[selects[0]])

func _on_open_recent_pressed() -> void:
	var selects := Array(%RecentList.get_selected_items())
	if selects.size():
		viewer.open_file(recent_file_paths[selects[0]])

func _ready() -> void:
	viewer.recent_file_paths_changed.connect(update_recent_file_paths)

